/*
 * Copyright (c) 1993 The Department of Computer Science, University of Calgary.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      Calgary, Department of Computer Science and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <sys/time.h>
#include <sys/procfs.h>

#include "syscall.h"

#ifndef MIN
#define	MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

/*
 *	This amount is added to the number in pr_syscall
 *	before the call is searched for in the syscalls[]
 *	table.  Essentially the difference between <sys.s>
 *	and reality.
 */
#define	DELTA	-1

#define hash(a,b,c,d)	(((u_long)(a) << 24) | \
			 ((u_long)(b) << 16) | \
			 ((u_long)(c) << 8) |  \
			 ((u_long)(d)))

static
char sccsid[] =
    "@(#)syscall.c 1.6 93/12/01 jda";

cmp(a, b)
long *a;
struct syscall *b;
{
	return(*a - b->num);
}

struct syscall *locate(n)
long n;
{
	struct syscall *p;

	n += DELTA;
	p = (struct syscall *)bsearch(&n, syscalls,
				      sizeof(syscalls)/sizeof(struct syscall),
				      sizeof(struct syscall), cmp);
	if (p)
		return(p);
	die("don't know what syscall %d is", n);
}

pr_callname(fp, p)
FILE *fp;
prstatus_t *p;
{
	extern int dash_t;
	struct syscall *sp = locate(p->pr_syscall);

	if (dash_t) {
		long t = time(NULL);
		fprintf(stderr, "%.24s: ", ctime(&t));
	}
	fprintf(fp, "%s(", sp->name);
}

pr_chr(fp, ch)
FILE *fp;
{
	if (ch < ' ')
		return(fprintf(fp, "^%c", ch + '@'));
	if (ch == 0177)
		return(fprintf(fp, "^?"));
	if (ch & 0x80)
		return(fprintf(fp, "%03o", ch));
	fputc(ch, fp);
}

pr_syscall(fd, fp, p)
FILE *fp;
prstatus_t *p;
{
	int i = 0;
	char *cp;
	struct syscall *sp = locate(p->pr_syscall);

	while (i < p->pr_nsysarg) {
		if (i >= sp->len)
			fprintf(fp, "0x%x", p->pr_sysarg[i]);
		else {
			cp = sysargs[sp->start + i];
			switch (hash(cp[0],cp[1],cp[2],cp[3])) {
			    case '%d\0\0':
				fprintf(fp, "%d", p->pr_sysarg[i]);
				break;

			    case '[%u]':
			    case '%u\0\0':
				fprintf(fp, "%u", p->pr_sysarg[i]);
				break;

			    case '[%o]':
			    case '%o\0\0':
				fprintf(fp, "0%02o", p->pr_sysarg[i]);
				break;

			    case '%sig':
				fprintf(fp, "%s", sigmap(p->pr_sysarg[i]));
				break;

			    case '%sgi':
				pr_syssgi(fp, p->pr_sysarg[i]);
				break;

			    case '%ctl':
				pr_ioctl(fp, p->pr_sysarg[i]);
				break;

			    case '%3s\0':
			    case '%s\0\0': {
				int j, nul, len;
				static char *buf;
				extern int dash_l;

				if (!buf) {
					buf = (char *)malloc(dash_l);
					if (!buf)
						die("out of memory");
				}

				nul = (strcmp("%s", cp) ? 0 : !0);
				len = (nul ? dash_l
					   : MIN(dash_l, p->pr_sysarg[2]));

				if (lseek(fd, p->pr_sysarg[i], 0) < 0 ||
				    read(fd, buf, len) != len)
				{
					fprintf(fp, "0x%x", p->pr_sysarg[i]);
					break;
				}
				fputc('\"', fp);
				for (j = 0; j < len; j++) {
					if (nul && !buf[j])
						break;
					pr_chr(fp, buf[j]);
				}
				if ((nul && j == len) ||
				    (!nul && j != p->pr_sysarg[2]))
				{
					fprintf(fp, "..");
				}
				fputc('\"', fp);
				break;
			    }

			    case '%x..':
			    case '[%x]':
			    case '%x\0\0':
				fprintf(fp, "0x%x", p->pr_sysarg[i]);
				break;
			    case '-\0\0\0':
				break;
			    default:
				die("unknown format type for %s\n", sp->name);
			}
		}

		if (++i < p->pr_nsysarg)
			fprintf(fp, ", ");
	}

	fprintf(fp, ") = ");
}

pr_sysret(fp, p)
FILE *fp;
prstatus_t *p;
{
	struct syscall *sp = locate(p->pr_syscall);
	char *cp = sysargs[sp->start + sp->len-1];

	if (p->pr_errno) {
		fprintf(fp, "-1 (%s)\n", strerror(p->pr_errno));
		return;
	}

	if (strncmp(cp, "->", 2))
		fprintf(fp, "%u\n", p->pr_rval1);
	else {
		switch (hash(cp[0],cp[1],cp[2],cp[3])) {
		    case '->%x':
			fprintf(fp, "0x%x\n", p->pr_rval1);
			break;
		    default:
			die("unknown return type for %s\n", sp->name);
		}
	}
}
